import json
import os

from libcoveocds.lib.common_checks import get_bad_ocid_prefixes, get_releases_aggregates
from tests import fixture_path

EMPTY_RELEASE_AGGREGATE = {
    "award_doc_count": 0,
    "award_doctype": {},
    "award_count": 0,
    "contract_doc_count": 0,
    "contract_doctype": {},
    "contract_count": 0,
    "contracts_without_awards": [],
    "duplicate_release_ids": [],
    "implementation_count": 0,
    "implementation_doc_count": 0,
    "implementation_doctype": {},
    "implementation_milestones_doc_count": 0,
    "implementation_milestones_doctype": {},
    "item_identifier_schemes": [],
    "max_award_date": "",
    "max_contract_date": "",
    "max_release_date": "",
    "max_tender_date": "",
    "min_award_date": "",
    "min_contract_date": "",
    "min_release_date": "",
    "min_tender_date": "",
    "organisations_with_address": 0,
    "organisations_with_contact_point": 0,
    "planning_doc_count": 0,
    "planning_doctype": {},
    "planning_count": 0,
    "award_item_count": 0,
    "contract_item_count": 0,
    "release_count": 0,
    "tender_item_count": 0,
    "tags": {},
    "tender_doc_count": 0,
    "tender_doctype": {},
    "tender_milestones_doc_count": 0,
    "tender_milestones_doctype": {},
    "tender_count": 0,
    "unique_award_id": [],
    "unique_buyers_count": 0,
    "unique_buyers_identifier": {},
    "unique_buyers_name_no_id": [],
    "unique_currency": [],
    "unique_initation_type": [],
    "unique_item_ids_count": 0,
    "unique_lang": [],
    "unique_ocids": [],
    "unique_org_count": 0,
    "unique_org_identifier_count": 0,
    "unique_org_name_count": 0,
    "unique_organisation_schemes": [],
    "unique_procuring_count": 0,
    "unique_procuring_identifier": {},
    "unique_procuring_name_no_id": [],
    "unique_suppliers_count": 0,
    "unique_suppliers_identifier": {},
    "unique_suppliers_name_no_id": [],
    "unique_tenderers_count": 0,
    "unique_tenderers_identifier": {},
    "unique_tenderers_name_no_id": [],
    "processes_implementation_count": 0,
    "processes_award_count": 0,
    "processes_contract_count": 0,
    "total_item_count": 0,
    "unique_buyers": [],
    "unique_procuring": [],
    "unique_suppliers": [],
    "unique_tenderers": [],
}

EXPECTED_RELEASE_AGGREGATE = {
    "award_count": 2,
    "award_doc_count": 3,
    "award_doctype": {"doctype1": 2, "doctype2": 1},
    "award_item_count": 2,
    "contract_count": 2,
    "contract_doc_count": 3,
    "contract_doctype": {"doctype1": 2, "doctype2": 1},
    "contract_item_count": 2,
    "contracts_without_awards": [
        {"awardID": "no", "id": "2", "period": {"startDate": "2015-01-02T00:00Z"}, "value": {"currency": "EUR"}}
    ],
    "duplicate_release_ids": [],
    "implementation_count": 1,
    "implementation_doc_count": 3,
    "implementation_doctype": {"doctype1": 2, "doctype2": 1},
    "implementation_milestones_doc_count": 3,
    "implementation_milestones_doctype": {"doctype1": 2, "doctype2": 1},
    "item_identifier_schemes": ["scheme1", "scheme2"],
    "max_award_date": "2015-01-02T00:00Z",
    "max_contract_date": "2015-01-02T00:00Z",
    "max_release_date": "2015-01-02T00:00Z",
    "max_tender_date": "2015-01-02T00:00Z",
    "min_award_date": "2015-01-01T00:00Z",
    "min_contract_date": "2015-01-01T00:00Z",
    "min_release_date": "2015-01-02T00:00Z",
    "min_tender_date": "2015-01-02T00:00Z",
    "organisations_with_address": 2,
    "organisations_with_contact_point": 2,
    "planning_count": 1,
    "planning_doc_count": 3,
    "planning_doctype": {"doctype1": 2, "doctype2": 1},
    "release_count": 1,
    "tags": {"planning": 1, "tender": 1},
    "tender_count": 1,
    "tender_doc_count": 3,
    "tender_doctype": {"doctype1": 2, "doctype2": 1},
    "tender_item_count": 2,
    "tender_milestones_doc_count": 3,
    "tender_milestones_doctype": {"doctype1": 2, "doctype2": 1},
    "unique_award_id": ["1", "2"],
    "unique_buyers_count": 1,
    "unique_buyers_identifier": {"1": "Gov"},
    "unique_buyers_name_no_id": [],
    "unique_currency": ["EUR", "GBP", "USD", "YEN"],
    "unique_initation_type": ["tender"],
    "unique_item_ids_count": 2,
    "unique_lang": ["English"],
    "unique_ocids": ["1"],
    "unique_org_count": 4,
    "unique_org_identifier_count": 2,
    "unique_org_name_count": 2,
    "unique_organisation_schemes": ["a", "b"],
    "unique_procuring_count": 1,
    "unique_procuring_identifier": {"1": "Gov"},
    "unique_procuring_name_no_id": [],
    "unique_suppliers_count": 3,
    "unique_suppliers_identifier": {"2": "Big corp3"},
    "unique_suppliers_name_no_id": ["Big corp1", "Big corp2"],
    "unique_tenderers_count": 3,
    "unique_tenderers_identifier": {"2": "Big corp3"},
    "unique_tenderers_name_no_id": ["Big corp1", "Big corp2"],
    "processes_implementation_count": 1,
    "processes_award_count": 1,
    "processes_contract_count": 1,
    "total_item_count": 6,
    "unique_buyers": ["Gov (1)"],
    "unique_procuring": ["Gov (1)"],
    "unique_suppliers": ["Big corp1", "Big corp2", "Big corp3 (2)"],
    "unique_tenderers": ["Big corp1", "Big corp2", "Big corp3 (2)"],
}

EXPECTED_RELEASE_AGGREGATE_RANDOM = {
    "award_count": 1477,
    "award_doc_count": 2159,
    "award_item_count": 2245,
    "contract_count": 1479,
    "contract_doc_count": 2186,
    "contract_item_count": 2093,
    "implementation_count": 562,
    "implementation_doc_count": 1140,
    "implementation_milestones_doc_count": 1659,
    "max_award_date": "5137-06-18T19:25:03.403Z",
    "max_contract_date": "5113-08-13T17:04:27.669Z",
    "max_release_date": "5137-06-07T19:44:36.152Z",
    "max_tender_date": "5084-05-16T03:11:56.723Z",
    "min_award_date": "1977-01-18T23:35:50.858Z",
    "min_contract_date": "1970-02-23T13:49:12.592Z",
    "min_release_date": "1970-08-03T00:21:37.491Z",
    "min_tender_date": "1989-03-20T22:29:02.697Z",
    "organisations_with_address": 770,
    "organisations_with_contact_point": 729,
    "planning_count": 339,
    "planning_doc_count": 738,
    "release_count": 1005,
    "tender_count": 467,
    "tender_doc_count": 799,
    "tender_item_count": 770,
    "tender_milestones_doc_count": 1163,
    "unique_buyers_count": 169,
    "unique_item_ids_count": 4698,
    "unique_org_count": 1339,
    "unique_org_identifier_count": 625,
    "unique_org_name_count": 714,
    "unique_procuring_count": 94,
    "unique_suppliers_count": 812,
    "unique_tenderers_count": 286,
    "processes_award_count": 466,
    "processes_contract_count": 467,
    "processes_implementation_count": 327,
    "total_item_count": 5108,
}


def test_get_releases_aggregates():
    assert get_releases_aggregates({}) == EMPTY_RELEASE_AGGREGATE
    assert get_releases_aggregates({"releases": []}) == EMPTY_RELEASE_AGGREGATE
    release_aggregate_3_empty = EMPTY_RELEASE_AGGREGATE.copy()
    release_aggregate_3_empty["release_count"] = 3
    assert get_releases_aggregates({"releases": [{}, {}, {}]}) == release_aggregate_3_empty

    with open(fixture_path("lib", "fixtures", "common_checks", "release_aggregate.json")) as fp:
        data = json.load(fp)

    assert get_releases_aggregates({"releases": data["releases"]}) == EXPECTED_RELEASE_AGGREGATE

    # test if a release is duplicated
    actual = get_releases_aggregates({"releases": data["releases"] + data["releases"]})
    actual_cleaned = {key: actual[key] for key in actual if "doc" not in key}
    actual_cleaned.pop("contracts_without_awards")

    expected_cleaned = {key: EXPECTED_RELEASE_AGGREGATE[key] for key in EXPECTED_RELEASE_AGGREGATE if "doc" not in key}
    expected_cleaned["tags"] = {"planning": 2, "tender": 2}
    expected_cleaned.pop("contracts_without_awards")
    expected_cleaned["release_count"] = 2
    expected_cleaned["duplicate_release_ids"] = ["1"]

    assert actual_cleaned == expected_cleaned

    with open(fixture_path("lib", "fixtures", "common_checks", "samplerubbish.json")) as fp:
        data = json.load(fp)

    actual = get_releases_aggregates(data)
    actual_cleaned = {key: actual[key] for key in actual if isinstance(actual[key], (str, int, float))}

    assert actual_cleaned == EXPECTED_RELEASE_AGGREGATE_RANDOM


def test_release_bad_ocid_prefixes():
    file_name = fixture_path("lib", "fixtures", "common_checks", "tenders_releases_7_releases_check_ocids.json")
    results = [
        ("bad-prefix-000001", "releases/0/ocid"),
        ("bad-prefix-000002", "releases/1/ocid"),
        ("bad-prefix-000002", "releases/2/ocid"),
        ("ocds-bad-000004", "releases/4/ocid"),
        ("ocds-bad-000004", "releases/5/ocid"),
        ("ocds-bad-000004", "releases/6/ocid"),
    ]

    with open(os.path.join(file_name)) as fp:
        user_data = json.load(fp)

    user_data_ocids = []
    for rel in user_data["releases"]:
        user_data_ocids.append(rel["ocid"])

    assert len(user_data_ocids) == 7  # 1 good, 6 bad ocds prefixes
    assert "ocds-00good-000003" in user_data_ocids  # good ocds prefix
    assert get_bad_ocid_prefixes(user_data) == results


def test_record_bad_ocid_prefixes_with_bad_compiled_release():
    file_name = fixture_path("lib", "fixtures", "common_checks", "record_check_ocids.json")
    results = [
        ("bad-prefix-000001", "records/0/ocid"),
    ]

    with open(os.path.join(file_name)) as fp:
        user_data = json.load(fp)

    assert get_bad_ocid_prefixes(user_data) == results



import os
import shutil
import tempfile

import pytest

import libcoveocds.config
from libcoveocds.api import APIException, ocds_json_output
from tests import fixture_path

# Cache for faster tests.
config = libcoveocds.config.LibCoveOCDSConfig()
config.config["cache_all_requests"] = True


def test_basic_1():

    cove_temp_folder = tempfile.mkdtemp(prefix="lib-cove-ocds-tests-", dir=tempfile.gettempdir())
    json_filename = fixture_path("fixtures", "api", "basic_1.json")

    results = ocds_json_output(
        cove_temp_folder, json_filename, schema_version="", convert=False, lib_cove_ocds_config=config
    )

    assert results["version_used"] == "1.1"
    assert results["validation_errors"] == []


def test_basic_record_package():

    cove_temp_folder = tempfile.mkdtemp(prefix="lib-cove-ocds-tests-", dir=tempfile.gettempdir())
    json_filename = fixture_path("fixtures", "api", "basic_record_package.json")

    results = ocds_json_output(
        cove_temp_folder, json_filename, schema_version="", convert=False, lib_cove_ocds_config=config, record_pkg=True
    )

    assert results["version_used"] == "1.1"
    assert results["validation_errors"] == []


@pytest.mark.parametrize(
    "json_data,expected",
    [
        ("{[,]}", "The file looks like invalid json"),
        (
            '{"version": "1.bad"}',
            "\x1b[1;31mThe schema version in your data is not valid. Accepted values: ['1.0', '1.1']\x1b[1;m",
        ),
    ],
)
def test_ocds_json_output_bad_data(json_data, expected):
    cove_temp_folder = tempfile.mkdtemp(prefix="lib-cove-ocds-tests-", dir=tempfile.gettempdir())

    file_path = os.path.join(cove_temp_folder, "bad_data.json")
    with open(file_path, "w") as fp:
        fp.write(json_data)
    try:
        with pytest.raises(APIException) as excinfo:
            ocds_json_output(cove_temp_folder, file_path, schema_version="", convert=False)

        assert str(excinfo.value) == expected
    finally:
        shutil.rmtree(cove_temp_folder)



import os
import warnings

from libcove.lib.tools import get_file_type

from libcoveocds.common_checks import common_checks_ocds
from libcoveocds.config import LibCoveOCDSConfig
from libcoveocds.lib.api import context_api_transform
from libcoveocds.schema import SchemaOCDS
from libcoveocds.util import json

try:
    from libcove.lib.common import get_spreadsheet_meta_data
    from libcove.lib.converters import convert_json, convert_spreadsheet
except ImportError:
    pass


class APIException(Exception):
    pass


def ocds_json_output(
    output_dir,
    file,
    schema_version,
    convert,
    file_type=None,
    json_data=None,
    lib_cove_ocds_config=None,
    record_pkg=None,
):
    """
    If flattentool is not installed, ``file_type`` must be ``"json"`` and ``convert`` must be falsy.
    """

    if not lib_cove_ocds_config:
        lib_cove_ocds_config = LibCoveOCDSConfig()
    if not file_type:
        file_type = get_file_type(file)
    if not json_data and file_type == "json":
        with open(file, "rb") as f:
            try:
                json_data = json.loads(f.read())
            except ValueError:
                raise APIException("The file looks like invalid json")

    if record_pkg is None:
        record_pkg = "records" in json_data

    if file_type == "json":
        schema_obj = SchemaOCDS(
            schema_version, json_data, lib_cove_ocds_config=lib_cove_ocds_config, record_pkg=record_pkg
        )
    else:
        metatab_schema_url = SchemaOCDS(select_version="1.1", lib_cove_ocds_config=lib_cove_ocds_config).pkg_schema_url
        metatab_data = get_spreadsheet_meta_data(output_dir, file, metatab_schema_url, file_type=file_type)
        schema_obj = SchemaOCDS(schema_version, lib_cove_ocds_config=lib_cove_ocds_config, package_data=metatab_data)

    if schema_obj.invalid_version_data:
        msg = "\033[1;31mThe schema version in your data is not valid. Accepted values: {}\033[1;m"
        raise APIException(msg.format(str(list(lib_cove_ocds_config.config["schema_version_choices"]))))

    if schema_obj.extensions:
        schema_obj.create_extended_schema_file(output_dir, "")

    # Used in conversions.
    schema_url = schema_obj.extended_schema_file or schema_obj.schema_url

    context = {"file_type": file_type}
    if file_type == "json":
        if convert:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")  # flattentool uses UserWarning, so we can't set a specific category

                context.update(
                    convert_json(
                        output_dir,
                        "",
                        file,
                        lib_cove_ocds_config,
                        schema_url=schema_url,
                        cache=False,
                        flatten=True,
                    )
                )
    else:
        context.update(
            convert_spreadsheet(
                output_dir,
                "",
                file,
                file_type,
                lib_cove_ocds_config,
                schema_url=schema_url,
                cache=False,
                pkg_schema_url=schema_obj.pkg_schema_url,
            )
        )

        with open(context["converted_path"], encoding="utf-8") as fp:
            json_data = json.load(fp)

    # context is edited in-place.
    context_api_transform(
        common_checks_ocds(
            context,
            output_dir,
            json_data,
            schema_obj,
            # `cache` writes the results to a file, which is only relevant in the context of repetitive web requests.
            cache=False,
        )
    )

    context["json_deref_error"] = schema_obj.json_deref_error
    context["invalid_version_data"] = schema_obj.invalid_version_data

    if file_type == "xlsx":
        # Remove unwanted files in the output
        # TODO: can we do this by no writing the files in the first place?
        os.remove(os.path.join(output_dir, "heading_source_map.json"))
        os.remove(os.path.join(output_dir, "cell_source_map.json"))

    return context



from libcoveocds.lib.api import context_api_transform


def test_context_api_transform_validation_additional_fields():
    context = {
        "validation_errors": [
            [
                '{"message_type":"type_a", "message":"description_a", "path_no_number":"field_a"}',
                [{"path": "path_to_a", "value": "a_value"}],
            ],
            [
                '{"message_type":"type_b", "message":"description_b", "path_no_number":"field_b"}',
                [{"path": "path_to_b", "value": ""}],
            ],
            [
                '{"message_type":"type_c", "message":"description_c", "path_no_number":"field_c"}',
                [{"path": "path_to_c"}],
            ],
        ],
        "data_only": [["path_to_d", "field_d", 1], ["path_to_e", "field_e", 2], ["path_to_f", "field_f", 3]],
        "additional_fields": {"path_to_d": {"some_info": "some_value"}},
        "additional_fields_count": 6,
        "validation_errors_count": 3,
        "deprecated_fields": {},
        "extensions": {},
    }
    expected_context = {
        "additional_fields": [
            {"field": "field_d", "path": "path_to_d", "usage_count": 1},
            {"field": "field_e", "path": "path_to_e", "usage_count": 2},
            {"field": "field_f", "path": "path_to_f", "usage_count": 3},
        ],
        "all_additional_fields": [{"some_info": "some_value"}],
        "deprecated_fields": [],
        "extensions": {},
        "validation_errors": [
            {
                "description": "description_a",
                "field": "field_a",
                "path": "path_to_a",
                "type": "type_a",
                "value": "a_value",
            },
            {
                "description": "description_b",
                "field": "field_b",
                "path": "path_to_b",
                "type": "type_b",
                "value": "",
            },
            {"description": "description_c", "field": "field_c", "path": "path_to_c", "type": "type_c", "value": ""},
        ],
    }
    transform_context = context_api_transform(context)

    validation_errors = zip(transform_context["validation_errors"], expected_context["validation_errors"])
    for result in validation_errors:
        assert len(result[0]) == len(result[1])
        for k, v in result[0].items():
            assert result[1][k] == v
    assert len(transform_context["validation_errors"]) == len(expected_context["validation_errors"])

    additional_fields = zip(transform_context["additional_fields"], expected_context["additional_fields"])
    for result in additional_fields:
        assert len(result[0]) == len(result[1])
        for k, v in result[0].items():
            assert result[1][k] == v
    assert len(transform_context["additional_fields"]) == len(expected_context["additional_fields"])

    assert transform_context["extensions"] == expected_context["extensions"]
    assert transform_context["deprecated_fields"] == expected_context["deprecated_fields"]
    assert "validation_errors_count" not in transform_context.keys()
    assert "additional_fields_count" not in transform_context.keys()
    assert "data_only" not in transform_context.keys()


def test_context_api_transform_extensions():
    """Expected result for extensions after transform:

    'extensions': {
        'extended_schema_url': 'extended_release_schema.json',
        'extensions': [
            {'description': 'description_a', 'documentationUrl': 'documentation_a', 'name': 'a', 'url': 'url_a'},
            {'description': 'description_b', 'documentationUrl': 'documentation_b', 'name': 'b', 'url': 'url_b'},
            {'description': 'description_c', 'documentationUrl': 'documentation_c', 'name': 'c', 'url': 'url_c'},
            {'description': 'description_d', 'documentationUrl': 'documentation_d', 'name': 'd', 'url': 'url_d'},
        ],
        'invalid_extensions': [['bad_url_x', 'x_error_msg'],
                               ['bad_url_z', 'z_error_msg'],
                               ['bad_url_y', 'y_error_msg']],
        'is_extended_schema': True
    }
    """
    context = {
        "extensions": {
            "is_extended_schema": True,
            "invalid_extension": {"bad_url_x": "x_error_msg", "bad_url_y": "y_error_msg", "bad_url_z": "z_error_msg"},
            "extensions": {
                "url_a": {
                    "description": "description_a",
                    "url": "url_a",
                    "documentationUrl": "documentation_a",
                    "name": "a",
                },
                "url_b": {
                    "description": "description_b",
                    "url": "url_b",
                    "documentationUrl": "documentation_b",
                    "name": "b",
                },
                "url_c": {
                    "description": "description_c",
                    "url": "url_c",
                    "documentationUrl": "documentation_c",
                    "name": "c",
                },
                "url_d": {
                    "description": "description_d",
                    "url": "url_d",
                    "documentationUrl": "documentation_d",
                    "name": "d",
                },
                "bad_url_x": [],
                "bad_url_y": [],
                "bad_url_z": [],
            },
            "extended_schema_url": "extended_release_schema.json",
        },
        "validation_errors_count": 0,
        "additional_fields_count": 0,
        "data_only": [],
        "validation_errors": [],
        "additional_fields": {},
        "deprecated_fields": {},
    }

    transformed_ext_context = context_api_transform(context)["extensions"]

    assert isinstance(transformed_ext_context["extensions"], list)
    assert len(transformed_ext_context["extensions"]) == 4
    for extension in transformed_ext_context["extensions"]:
        assert len(extension) == 4

    assert len(transformed_ext_context["invalid_extensions"]) == 3
    for inv_extension in transformed_ext_context["invalid_extensions"]:
        assert len(inv_extension) == 2
        for extension in transformed_ext_context["extensions"]:
            assert extension["url"] != inv_extension[0]

    assert transformed_ext_context["is_extended_schema"] == context["extensions"]["is_extended_schema"]
    assert transformed_ext_context["extended_schema_url"] == context["extensions"]["extended_schema_url"]


def test_context_api_transform_deprecations():
    """Expected result for deprecated field after transform:

    'deprecated_fields': [
        {"field": "a", "paths": ["path_to_a/0/a", "path_to_a/1/a"], "explanation": ["1.1", "description_a"]},
        {"field": "b", "paths": ["path_to_b/0/b", "path_to_b/1/b"], "explanation": ["1.1", "description_b"]}
    ]
    """
    context = {
        "deprecated_fields": {
            "a": {"paths": ["path_to_a/0/a", "path_to_a/1/a"], "explanation": ["1.1", "description_a"]},
            "b": {"paths": ["path_to_b/0/b", "path_to_b/1/b"], "explanation": ["1.1", "description_b"]},
        },
        "validation_errors": [],
        "validation_errors_count": 0,
        "data_only": [],
        "additional_fields": {},
        "additional_fields_count": 0,
        "extensions": {},
    }

    transformed_depr_context = context_api_transform(context)["deprecated_fields"]

    assert isinstance(transformed_depr_context, list)
    assert len(transformed_depr_context) == 2

    for deprecated_field in transformed_depr_context:
        assert isinstance(deprecated_field, dict)
        assert len(deprecated_field) == 3
        assert len(deprecated_field["paths"]) == 2
        assert len(deprecated_field["explanation"]) == 2
        assert deprecated_field.get("field")



LIB_COVE_OCDS_CONFIG_DEFAULT = {
    # SchemaOCDS options
    # Note: "schema_version" is set after this dict.
    #
    # Used indirectly by lib-cove for "version_display_choices" and "version_used_display".
    "schema_version_choices": {
        # version: (display, url, tag),
        "1.0": ("1.0", "https://standard.open-contracting.org/1.0/en/", "1__0__3"),
        "1.1": ("1.1", "https://standard.open-contracting.org/1.1/en/", "1__1__5"),
    },
    # Used indirectly by lib-cove to construct "codelist_url".
    "schema_codelists": {
        # version: codelist_dir,
        "1.1": "https://raw.githubusercontent.com/open-contracting/standard/1.1/schema/codelists/",
    },
    # The language key to use to read extension metadata.
    "current_language": "en",
    # Path to ZIP file of standard repository.
    "standard_zip": None,
    #
    # lib-cove-web options
    #
    "app_name": "cove_ocds",
    "app_base_template": "cove_ocds/base.html",
    "app_verbose_name": "Open Contracting Data Review Tool",
    "app_strapline": "Review your OCDS data.",
    "input_methods": ["upload", "url", "text"],
    "support_email": "data@open-contracting.org",
    #
    # Flatten Tool options
    #
    "root_list_path": "releases",
    "root_id": "ocid",
    "convert_titles": False,
    "flatten_tool": {
        "disable_local_refs": True,
        "remove_empty_schema_columns": True,
    },
    #
    # lib-cove-ocds options
    #
    # Whether to cache requests that pass through to lib-cove.
    "cache_all_requests": False,
    # Which additional checks to perform ("all" or "none", per libcoveocds.lib.additional_checks.CHECKS).
    "additional_checks": "all",
    # Whether to add "releases_aggregates" and "records_aggregates" to the context.
    "skip_aggregates": False,
    # The context in which lib-cove-ocds is used ("web" or "api").
    "context": "web",
}

# Set default schema version to the latest version
LIB_COVE_OCDS_CONFIG_DEFAULT["schema_version"] = list(LIB_COVE_OCDS_CONFIG_DEFAULT["schema_version_choices"])[-1]


class LibCoveOCDSConfig:
    def __init__(self, config=None):
        # We need to make sure we take a copy,
        #   so that changes to one config object don't end up effecting other config objects.
        if config:
            self.config = config.copy()
        else:
            self.config = LIB_COVE_OCDS_CONFIG_DEFAULT.copy()



import json
import os
import shutil
import tempfile

from click.testing import CliRunner

from libcoveocds.__main__ import main


def test_basic():
    runner = CliRunner()
    result = runner.invoke(main, [os.path.join("tests", "fixtures", "common_checks", "basic_1.json")])
    data = json.loads(result.output)

    assert result.exit_code == 0
    assert "1.1" == data.get("version_used")


def test_old_schema():
    runner = CliRunner()
    result = runner.invoke(main, ["-s", "1.0", os.path.join("tests", "fixtures", "common_checks", "basic_1.json")])

    assert result.exit_code == 0

    data = json.loads(result.output)

    assert "1.0" == data.get("version_used")


def test_set_output_dir():
    output_dir = tempfile.mkdtemp(
        prefix="lib-cove-ocds-tests-",
        dir=tempfile.gettempdir(),
    )
    runner = CliRunner()
    result = runner.invoke(
        main, ["-o", output_dir, os.path.join("tests", "fixtures", "common_checks", "basic_1.json")]
    )

    # This will fail because tempfile.mkdtemp already will make the directory, and so it already exists
    assert result.exit_code == 1
    assert result.output.startswith("Directory ")
    assert result.output.endswith("already exists\n")

    shutil.rmtree(output_dir)


def test_set_output_dir_and_delete():
    output_dir = tempfile.mkdtemp(
        prefix="lib-cove-ocds-tests-",
        dir=tempfile.gettempdir(),
    )
    runner = CliRunner()
    result = runner.invoke(
        main, ["-d", "-o", output_dir, os.path.join("tests", "fixtures", "common_checks", "basic_1.json")]
    )
    # Should have results file and original file and nothing else
    expected_files = ["basic_1.json", "results.json"]

    assert result.exit_code == 0

    data = json.loads(result.output)

    assert "1.1" == data.get("version_used")
    assert sorted(os.listdir(output_dir)) == sorted(expected_files)

    shutil.rmtree(output_dir)


def test_set_output_dir_and_delete_and_exclude():
    output_dir = tempfile.mkdtemp(
        prefix="lib-cove-ocds-tests-",
        dir=tempfile.gettempdir(),
    )
    runner = CliRunner()
    result = runner.invoke(
        main, ["-d", "-e", "-o", output_dir, os.path.join("tests", "fixtures", "common_checks", "basic_1.json")]
    )
    # Should have results file only
    expected_files = ["results.json"]

    assert result.exit_code == 0

    data = json.loads(result.output)

    assert "1.1" == data.get("version_used")
    assert sorted(os.listdir(output_dir)) == sorted(expected_files)

    shutil.rmtree(output_dir)


def test_set_output_dir_and_convert():
    output_dir = tempfile.mkdtemp(
        prefix="lib-cove-ocds-tests-",
        dir=tempfile.gettempdir(),
    )
    runner = CliRunner()
    result = runner.invoke(
        main, ["-c", "-d", "-o", output_dir, os.path.join("tests", "fixtures", "common_checks", "basic_1.json")]
    )
    # Should have results file and original file and the converted files
    expected_files = ["basic_1.json", "flattened", "flattened.ods", "flattened.xlsx", "results.json"]

    assert result.exit_code == 0

    data = json.loads(result.output)

    assert "1.1" == data.get("version_used")
    assert sorted(os.listdir(output_dir)) == sorted(expected_files)
    # Flattened should be a directory of csv's.
    # We aren't going to check names fully
    # That would leave this test brittle if flatten-tools naming scheme changed,
    # so we will just check extension
    for filename in os.listdir(os.path.join(output_dir, "flattened")):
        assert filename.endswith(".csv")

    shutil.rmtree(output_dir)



import json
import re

from jsonschema.exceptions import ValidationError, _RefResolutionError
from libcove.lib.common import common_checks_context, get_additional_codelist_values, unique_ids, validator
from libcove.lib.tools import decimal_default
from referencing.exceptions import Unresolvable

from libcoveocds.lib.additional_checks import CHECKS, run_additional_checks
from libcoveocds.lib.common_checks import get_bad_ocid_prefixes, get_records_aggregates, get_releases_aggregates

try:
    import bleach
    from django.utils.html import conditional_escape, escape, format_html, mark_safe
    from markdown_it import MarkdownIt

    md = MarkdownIt()

    validation_error_lookup = {
        "date-time": mark_safe(
            "Incorrect date format. Dates should use the form YYYY-MM-DDT00:00:00Z. Learn more about "
            '<a href="https://standard.open-contracting.org/latest/en/schema/reference/#date">dates in OCDS</a>.'
        ),
    }
except ImportError:
    pass


def unique_ids_or_ocids(validator, ui, instance, schema):
    # `records` key from the JSON schema doesn't get passed through to here, so
    # we look out for this $ref — this may change if the way the schema files
    # are structured changes.
    if schema.get("items") == {"$ref": "#/definitions/record"}:
        return unique_ids(validator, ui, instance, schema, id_names=["ocid"])
    elif "$ref" in schema.get("items", {}) and schema["items"]["$ref"].endswith("release-schema.json"):
        return unique_ids(validator, ui, instance, schema, id_names=["ocid", "id"])
    else:
        return unique_ids(validator, ui, instance, schema, id_names=["id"])


def oneOf_draft4(validator, oneOf, instance, schema):
    """
    oneOf_draft4 validator from
    https://github.com/Julian/jsonschema/blob/d16713a4296663f3d62c50b9f9a2893cb380b7af/jsonschema/_validators.py#L337

    Modified to:
    - sort the instance JSON, so we get a reproducible output that we
      can can test more easily
    - Yield all the individual errors for linked or embedded releases within a
      record.
    - Return more information on the ValidationError object, to allow us to
      replace the translation with a message in cove-ocds
    """
    subschemas = enumerate(oneOf)
    all_errors = []
    for index, subschema in subschemas:
        errs = list(validator.descend(instance, subschema, schema_path=index))
        if not errs:
            first_valid = subschema
            break
        # We check the title, because we don't have access to the field name,
        # as it lives in the parent.
        # It will not match the releases array in a release package, because
        # there is no oneOf.
        if (
            schema.get("title") == "Releases"
            or schema.get("description") == "An array of linking identifiers or releases"
        ):
            # If instance is not a list, or is a list of zero length, then
            # validating against either subschema will work.
            # Assume instance is an array of Linked releases, if there are no
            # "id"s in any of the releases.
            if type(instance) is not list or all("id" not in release for release in instance):
                if "properties" in subschema.get("items", {}) and "id" not in subschema["items"]["properties"]:
                    for err in errs:
                        err.assumption = "linked_releases"
                        yield err
                    return
            # Assume instance is an array of Embedded releases, if there is an
            # "id" in each of the releases
            elif all("id" in release for release in instance):
                if "id" in subschema.get("items", {}).get("properties", {}) or subschema.get("items", {}).get(
                    "$ref", ""
                ).endswith("release-schema.json"):
                    for err in errs:
                        err.assumption = "embedded_releases"
                        yield err
                    return
            else:
                err = ValidationError(
                    "This array should contain either entirely embedded releases or "
                    "linked releases. Embedded releases contain an 'id' whereas linked "
                    "releases do not. Your releases contain a mixture."
                )
                err.error_id = "releases_both_embedded_and_linked"
                yield err
                return

        all_errors.extend(errs)
    else:
        err = ValidationError(
            f"{json.dumps(instance, sort_keys=True, default=decimal_default)} "
            "is not valid under any of the given schemas",
            context=all_errors,
        )
        err.error_id = "oneOf_any"
        yield err

    more_valid = [s for i, s in subschemas if validator.evolve(schema=s).is_valid(instance)]
    if more_valid:
        more_valid.append(first_valid)
        reprs = ", ".join(repr(schema) for schema in more_valid)
        err = ValidationError(f"{instance!r} is valid under each of {reprs}")
        err.error_id = "oneOf_each"
        err.reprs = reprs
        yield err


validator.VALIDATORS["uniqueItems"] = unique_ids_or_ocids
validator.VALIDATORS["oneOf"] = oneOf_draft4


# ref_into is used calculate the HTML anchor for the field in the OCDS documentation.
def _lookup_schema(schema, path, ref_info=None):
    if not path:
        return schema, ref_info

    if hasattr(schema, "__reference__"):
        ref_info = {"path": path, "reference": schema.__reference__}

    if "items" in schema:
        return _lookup_schema(schema["items"], path, ref_info)
    if "properties" in schema:
        head, *tail = path
        if head in schema["properties"]:
            return _lookup_schema(schema["properties"][head], tail, ref_info)
    return None, None


def common_checks_ocds(
    context,
    upload_dir,
    json_data,
    schema_obj,
    cache=True,
):
    """
    param skip_aggregates: whether to skip "releases_aggregates" and "records_aggregates"
    """
    skip_aggregates = schema_obj.config.config["skip_aggregates"]
    additional_checks = CHECKS[schema_obj.config.config["additional_checks"]]

    # Pass "-" as the schema name. The associated logic is not required by lib-cove-ocds.
    try:
        common_checks = common_checks_context(
            upload_dir, json_data, schema_obj, "-", context, fields_regex=True, api=schema_obj.api, cache=cache
        )
    except (Unresolvable, _RefResolutionError) as e:
        # "PointerToNowhere: '/definitions/Unresolvable' does not exist within {big JSON blob}"
        schema_obj.json_deref_error = re.sub(r" within .+", "", str(e))
        return context

    ignore_errors = bool(common_checks["context"]["validation_errors"])

    # Note: Pelican checks whether the OCID prefix is registered.
    ocds_prefixes_bad_format = get_bad_ocid_prefixes(json_data)
    if ocds_prefixes_bad_format:
        context["conformance_errors"] = {"ocds_prefixes_bad_format": ocds_prefixes_bad_format}

    # If called in an API context:
    # - Skip the schema description and reference URL for OCID prefix conformance errors.
    # - Skip the formatted message, schema title, schema description and reference URL for validation errors.
    if not schema_obj.api:
        if "conformance_errors" in context:
            ocid_description = schema_obj.get_schema_obj()["properties"]["ocid"]["description"]
            # XXX: The last sentence is assumed to be a link to guidance in all versions of OCDS.
            index = ocid_description.rindex(". ") + 1
            context["conformance_errors"]["ocid_description"] = ocid_description[:index]
            context["conformance_errors"]["ocid_info_url"] = re.search(r"\((\S+)\)", ocid_description[index:]).group(1)

        new_validation_errors = []
        for (json_key, values) in common_checks["context"]["validation_errors"]:
            error = json.loads(json_key)

            new_message = validation_error_lookup.get(error["message_type"])
            if new_message:
                error["message_safe"] = conditional_escape(new_message)
            else:
                if "message_safe" in error:
                    error["message_safe"] = mark_safe(error["message_safe"])
                else:
                    error["message_safe"] = conditional_escape(error["message"])

            schema_block, ref_info = _lookup_schema(
                schema_obj.get_pkg_schema_obj(deref=True), error["path_no_number"].split("/")
            )
            if schema_block and error["message_type"] != "required":
                if "description" in schema_block:
                    error["schema_title"] = escape(schema_block.get("title", ""))
                    error["schema_description_safe"] = mark_safe(
                        bleach.clean(
                            md.render(schema_block["description"]), tags=bleach.sanitizer.ALLOWED_TAGS | {"p"}
                        )
                    )
                if ref_info:
                    ref = ref_info["reference"]["$ref"]
                    if ref.endswith("release-schema.json"):
                        ref = ""
                    else:
                        ref = ref.strip("#")
                    ref_path = "/".join(ref_info["path"])
                    schema = "release-schema.json"
                else:
                    ref = ""
                    ref_path = error["path_no_number"]
                    schema = "release-package-schema.json"
                error["docs_ref"] = format_html("{},{},{}", schema, ref, ref_path)

            new_validation_errors.append([json.dumps(error, sort_keys=True), values])
        common_checks["context"]["validation_errors"] = new_validation_errors

    context.update(common_checks["context"])

    if not skip_aggregates:
        if "records" in json_data:
            context["records_aggregates"] = get_records_aggregates(json_data, ignore_errors=ignore_errors)
        else:
            context["releases_aggregates"] = get_releases_aggregates(json_data, ignore_errors=ignore_errors)

    additional_codelist_values = get_additional_codelist_values(schema_obj, json_data)
    context["additional_closed_codelist_values"] = {
        key: value for key, value in additional_codelist_values.items() if not value["isopen"]
    }
    context["additional_open_codelist_values"] = {
        key: value for key, value in additional_codelist_values.items() if value["isopen"]
    }

    if additional_checks:
        context["additional_checks"] = run_additional_checks(
            json_data, additional_checks, ignore_errors=True, return_on_error=[]
        )

    return context



import json
import os
import shutil
import sys
import tempfile

import click

import libcoveocds.api
from libcoveocds.config import LibCoveOCDSConfig
from libcoveocds.lib.additional_checks import CHECKS


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


@click.command()
@click.argument("filename")
@click.option("-c", "--convert", is_flag=True, help="Convert data from nested (json) to flat format (spreadsheet)")
@click.option(
    "-o", "--output-dir", default=None, help="Directory where the output is created, defaults to the name of the file"
)
@click.option(
    "-s",
    "--schema-version",
    type=click.Choice(LibCoveOCDSConfig().config["schema_version_choices"]),
    help="Version of the schema to validate the data, eg '1.0'",
)
@click.option("-d", "--delete", is_flag=True, help="Delete existing directory if it exits")
@click.option("-e", "--exclude-file", is_flag=True, help="Do not include the file in the output directory")
@click.option(
    "--additional-checks", default="all", type=click.Choice(CHECKS), help="The set of additional checks to perform"
)
@click.option("--skip-aggregates", is_flag=True, help="Skip releases_aggregates and records_aggregates")
@click.option(
    "--standard-zip",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to a ZIP file containing the standard repository",
)
def main(
    filename,
    output_dir,
    convert,
    schema_version,
    delete,
    exclude_file,
    additional_checks,
    skip_aggregates,
    standard_zip,
):
    if standard_zip:
        standard_zip = f"file://{standard_zip}"

    config = LibCoveOCDSConfig()
    config.config["standard_zip"] = standard_zip
    config.config["cache_all_requests"] = True
    config.config["additional_checks"] = additional_checks
    config.config["skip_aggregates"] = skip_aggregates
    config.config["context"] = "api"

    if schema_version:
        version_choices = config.config.get("schema_version_choices")
        if schema_version not in version_choices:
            print(f"Value for schema version option is not valid. Accepted values: {', '.join(version_choices)}")
            sys.exit(1)

    # Do we have output on disk? We only do in certain modes
    has_disk_output = output_dir or convert or delete or exclude_file
    if has_disk_output:
        if not output_dir:
            output_dir = filename.split("/")[-1].split(".")[0]

        if os.path.exists(output_dir):
            if delete:
                shutil.rmtree(output_dir)
            else:
                print(f"Directory {output_dir} already exists")
                sys.exit(1)
        os.makedirs(output_dir)

        if not exclude_file:
            shutil.copy2(filename, output_dir)
    else:
        # If not, just put in /tmp and delete after
        output_dir = tempfile.mkdtemp(
            prefix="lib-cove-ocds-cli-",
            dir=tempfile.gettempdir(),
        )

    try:
        result = libcoveocds.api.ocds_json_output(
            output_dir, filename, schema_version, convert=convert, file_type="json", lib_cove_ocds_config=config
        )
    finally:
        if not has_disk_output:
            shutil.rmtree(output_dir)

    output = json.dumps(result, indent=2, cls=SetEncoder)
    if has_disk_output:
        with open(os.path.join(output_dir, "results.json"), "w") as fp:
            fp.write(output)

    print(output)


if __name__ == "__main__":
    main()



import json


def context_api_transform(context):
    """
    Reformat `context` for use in an API context.

    -  Remove ``validation_errors_count``
    -  Remove ``additional_fields_count``
    -  Reformat ``additional_fields`` from a dict to a list of its values
    -  Rename ``additional_fields`` to ``all_additional_fields``
    -  Reformat ``data_only`` from a 3-value tuple to a 3-key dict
    -  Rename ``data_only`` to ``additional_fields``
    -  Reformat ``deprecated_fields`` from a dict to a list of its values, with keys added to values as ``field`` keys
    -  Reformat ``validation_errors`` from [ [ "{..}", [ {..}, ..] ], ..] to [ {..}, ..]:

       -  ``type`` (``message_type`` from the JSON text, repeated for each item in the list of dicts)
       -  ``field`` (``path_no_number`` from the JSON text, repeated for each item in the list of dicts)
       -  ``description`` (``message`` from the JSON text, repeated for each item in the list of dicts)
       -  ``path`` from the list of dicts
       -  ``value`` from the list of dicts
       -  Ignore other keys

    -  Reformat ``extensions``:

       -  Reformat ``extensions`` from a dict to a list of its values, for the keys not in ``invalid_extensions``
       -  Reformat ``invalid_extensions`` from a dict to a list of 2-value lists
    """
    context.pop("validation_errors_count")
    context.pop("additional_fields_count")

    context["deprecated_fields"] = [{"field": key, **value} for key, value in context["deprecated_fields"].items()]
    context["all_additional_fields"] = list(context.pop("additional_fields").values())

    # Don't move this above `context.pop("additional_fields")`.
    context["additional_fields"] = [
        {"path": path, "field": field, "usage_count": usage_count}
        for path, field, usage_count in context.pop("data_only")
    ]

    validation_errors = []
    for error, path_values in context["validation_errors"]:
        error = json.loads(error)
        validation_errors.extend(
            {
                "type": error["message_type"],
                "field": error["path_no_number"],
                "description": error["message"],
                "path": path_value.get("path", ""),
                "value": path_value.get("value", ""),
            }
            for path_value in path_values
        )
    context["validation_errors"] = validation_errors

    extensions = context["extensions"]
    if extensions:
        invalid_extension = extensions["invalid_extension"]
        extensions["extensions"] = [
            value for key, value in extensions["extensions"].items() if key not in invalid_extension
        ]
        extensions["invalid_extensions"] = list(map(list, invalid_extension.items()))

    # Context is edited in-place, but existing callers might expect a return value.
    return context



import functools
import json
import logging
import os
import warnings
import zipfile
from collections import defaultdict
from copy import deepcopy
from urllib.parse import urljoin

import jsonref
import requests
from libcove.lib.common import SchemaJsonMixin, get_schema_codelist_paths
from ocdsextensionregistry.exceptions import DoesNotExist, ExtensionCodelistWarning, ExtensionWarning
from ocdsextensionregistry.profile_builder import ProfileBuilder
from referencing import Registry, Resource

import libcoveocds.config

logger = logging.getLogger(__name__)


class SchemaOCDS(SchemaJsonMixin):
    def _set_schema_version(self, version):
        schema_version_choice = self.version_choices[version]

        # lib-cove uses version in common_checks_context() via getattr() to determine whether to set
        # "version_display_choices" and "version_used_display".
        # cove-ocds uses version, e.g. when a user changes the version to validate against.
        self.version = version
        # lib-cove uses schema_host in get_schema_validation_errors() to call CustomRefResolver(). (URL, not host.)
        # The rationale was to allow overriding $ref in schema files. (It would've been easier to edit the schema....)
        # https://github.com/OpenDataServices/cove/commit/ccc46fffe430868a1cb0ddba356c719bb62896b5
        self.schema_host = schema_version_choice[1]
        self._schema_tag = schema_version_choice[2]

    def __init__(self, select_version=None, package_data=None, lib_cove_ocds_config=None, record_pkg=False):
        """Build the schema object using an specific OCDS schema version

        The version used will be select_version, package_data.get('version') or
        default version, in that order. Invalid version choices in select_version or
        package_data will be skipped and registered as self.invalid_version_argument
        and self.invalid_version_data respectively.
        """
        # The main configuration object.
        self.config = lib_cove_ocds_config or libcoveocds.config.LibCoveOCDSConfig()
        # Whether used in an API context.
        self.api = self.config.config["context"] == "api"

        # lib-cove uses schema_name in get_schema_validation_errors() to call CustomRefResolver(), if extended is set.
        # It does this to resolve the release schema $ref to the patched release schema. Although this level of
        # indirection is undesirable, there is no workaround in lib-cove<=0.31.0.
        self.schema_name = "release-schema.json"

        # lib-cove uses codelists in get_additional_codelist_values() for "codelist_url".
        self.codelists = self.config.config["schema_codelists"]["1.1"]
        # lib-cove uses version_choices in common_checks_context() via getattr() for "version_display_choices" and
        # "version_used_display". Re-used in this file for convenience.
        self.version_choices = self.config.config["schema_version_choices"]

        # Report errors in web UI.
        self.missing_package = False
        self.invalid_version_argument = False
        self.invalid_version_data = False
        self.json_deref_error = None  # str

        self._set_schema_version(self.config.config["schema_version"])

        # The selected version overrides the default version and the data version.
        if select_version:
            if select_version in self.version_choices:
                self._set_schema_version(select_version)
            elif self.api:
                raise AttributeError(
                    f"select_version: {select_version} is not one of {', '.join(self.version_choices)}"
                )
            else:
                self.invalid_version_argument = True
                # If invalid, use other strategies.
                select_version = None

        # lib-cove uses extensions in common_checks_context() for "extensions"."extensions".
        # If `self.extensions` is falsy, this logic can be skipped.
        # cove-ocds uses extensions to render extension-related information.
        self.extensions = {}
        if isinstance(package_data, dict):
            if "releases" not in package_data and "records" not in package_data:
                self.missing_package = True
            if not select_version:
                package_version = package_data.get("version")
                if package_version:
                    if package_version in self.version_choices:
                        self._set_schema_version(package_version)
                    else:
                        self.invalid_version_data = True

            extensions = package_data.get("extensions")
            if isinstance(extensions, list):
                self.extensions = {extension: {} for extension in extensions if isinstance(extension, str)}

        # (lib-)cove-ocds uses schema_url as a fallback to extended_schema_file, to convert between formats.
        self.schema_url = urljoin(self.schema_host, "release-schema.json")
        # Used in get_pkg_schema_obj() to determine which package to return.
        if record_pkg:
            self._package_schema_name = "record-package-schema.json"
        else:
            self._package_schema_name = "release-package-schema.json"
        # lib-cove uses pkg_schema_url in common_checks_context() for "schema_url".
        self.pkg_schema_url = urljoin(self.schema_host, self._package_schema_name)

        #: The profile builder instance for this package's extensions.
        self.builder = ProfileBuilder(
            self._schema_tag, list(self.extensions), standard_base_url=self.config.config["standard_zip"]
        )
        # Initialize extensions once and preserve locale caches.
        self.builder_extensions = list(self.builder.extensions())

        # lib-cove uses extended in common_checks_context() for "extensions"."is_extended_schema".
        # If `self.extensions` is falsy, this logic can be skipped.
        # cove-ocds uses is_extended_schema to conditionally display content.
        self.extended = False
        # lib-cove uses invalid_extension in common_checks_context() for "extensions"."invalid_extension".
        # If `self.extensions` is falsy, this logic can be skipped.
        # cove-ocds uses invalid_extension to render extension-related errors.
        self.invalid_extension = {}

        # lib-cove uses extended_schema_url in common_checks_context() for "extensions"."extended_schema_url".
        # If `self.extensions` is falsy, this logic can be skipped.
        # cove-ocds uses extended_schema_url to "Get a copy of the schema with extension patches applied".
        self.extended_schema_url = None
        # lib-cove uses extended_schema_file in get_schema_validation_errors() to call CustomRefResolver(), if extended
        # is set. (lib-)cove-ocds uses extended_schema_file to convert between formats, and falls back to schema_url.
        self.extended_schema_file = None

    def _codelist_codes(self, codelist):
        if not codelist.rows:
            return set()
        column = next(column for column in ("Code", "Código", "code") if column in codelist.rows[0])
        return {row[column] for row in codelist.rows}

    @functools.lru_cache
    def _standard_codelists(self):
        try:
            # OCDS 1.0 uses "code" column. https://github.com/open-contracting/standard/blob/1__0__3/standard/schema/codelists/organizationIdentifierRegistrationAgency_iati.csv  # noqa: 501
            return {codelist.name: self._codelist_codes(codelist) for codelist in self.builder.standard_codelists()}
        except requests.RequestException as e:
            logger.exception(e)
            return {}

    # lib-cove calls this from get_additional_codelist_values().
    def process_codelists(self):
        # lib-cove uses these in get_additional_codelist_values().
        # - Used to determine whether a field has a codelist, which codelist and whether it is open.
        self.extended_codelist_schema_paths = get_schema_codelist_paths(self, use_extensions=True)
        # - Used with the `in` operator, to determine whether a codelist is from an extension.
        self.core_codelists = self._standard_codelists()
        # - Used with get(), and the return value is used with `in`, to determine whether a code is included.
        self.extended_codelists = deepcopy(self.core_codelists)
        # - Used to populate "codelist_url" and "codelist_amend_urls".
        self.extended_codelist_urls = defaultdict(list)

        # _standard_codelists() returns an empty dict on HTTP error. If so, don't cache this empty dict, and return.
        if not self.core_codelists:
            self._standard_codelists.cache_clear()
            return

        for extension in self.builder_extensions:
            input_url = extension.input_url

            failed_codelists = self.extensions[input_url].get("failed_codelists")

            # In a web context, skip extensions whose metadata is unavailable.
            if failed_codelists is None:
                continue

            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always", category=ExtensionCodelistWarning)

                try:
                    # An unreadable metadata file or a malformed extension URL raises an error.
                    extension_codelists = extension.codelists
                except (requests.RequestException, NotImplementedError):
                    # patched_release_schema() will have recorded the metadata file being unreadable.
                    continue

                for warning in w:
                    if issubclass(warning.category, ExtensionCodelistWarning):
                        exception = warning.message.exc
                        if isinstance(exception, requests.HTTPError):
                            message = f"{exception.response.status_code}: {exception.response.reason}"
                        elif isinstance(exception, (requests.RequestException, zipfile.BadZipFile)):
                            message = "Couldn't be retrieved"
                        elif isinstance(exception, UnicodeDecodeError):
                            message = "Has non-UTF-8 characters"
                        else:
                            message = f"Unknown error: {exception}"
                        failed_codelists[warning.message.codelist] = message

            for name, codelist in extension_codelists.items():
                try:
                    codes = self._codelist_codes(codelist)
                except StopIteration:
                    failed_codelists[name] = 'Has no "Code" column'
                    continue

                if codelist.patch:
                    basename = codelist.basename

                    if basename not in self.core_codelists:
                        failed_codelists[name] = f"References non-existing codelist {basename}"
                        continue

                    patched_codelist = self.extended_codelists[basename]

                    if codelist.addend:
                        patched_codelist |= codes
                    elif codelist.subtrahend:
                        nonexisting_codes = [code for code in codes if code not in patched_codelist]
                        if nonexisting_codes:
                            failed_codelists[name] = f"References non-existing code(s): {', '.join(nonexisting_codes)}"
                        patched_codelist -= codes
                else:
                    self.extended_codelists[name] = codes

                self.extended_codelist_urls[name].append(extension.get_url(f"codelists/{name}"))

    @property
    def registry(self):
        # lib-cove's get_schema_validation_errors() expects a non-dereferenced schema.
        # The cove-ocds test with tenders_releases_1_release_with_extension_broken_json_ref.json fails, otherwise.
        release_schema = Resource.from_contents(self.get_schema_obj())
        versioned_release_schema = Resource.from_contents(
            json.loads(self.builder.get_standard_file_contents("versioned-release-validation-schema.json"))
        )
        return Registry().with_resources(
            [
                (f"{scheme}://standard.open-contracting.org/schema/{self._schema_tag}/{prefix}-schema.json", schema)
                # OCDS 1.0 use the http scheme.
                for scheme in ("http", "https")
                for prefix, schema in (
                    ("release", release_schema),
                    ("versioned-release-validation", versioned_release_schema),
                )
            ]
        )

    def _get_schema(self, name):
        # The ocds_babel.translate.translate() makes these substitutions for published files.
        return json.loads(
            self.builder.get_standard_file_contents(name)
            .replace("{{lang}}", self.config.config["current_language"])
            .replace("{{version}}", self.version)
        )

    # lib-cove calls this from many locations including get_schema_validation_errors().
    # use_extensions is always called as True, so it is ignored.
    #
    # ProfileBuilder.release_package_schema() and record_package_schema() aren't used, because the patched release
    # schema has already been calculated and cached by patched_release_schema().
    #
    # lib-cove resolves the release schema $ref to the patched release schema by overridding jsonschema's resolver. For
    # this to work, lib-cove needs create_extended_schema_file() to have been called before the validator runs, like
    # libcoveocds.api.ocds_json_output() and cove-ocds do.
    #
    # The resolver converts the $ref to its basename, making it impossible to circumvent this logic. It would have been
    # less indirect to set the $ref to the file written by create_extended_schema_file() in this method rather than via
    # the resolver.
    #
    # If using jsonschema>=4.18 and lib-cove#123, self.registry is used to achieve this.
    #
    # For reference, the old code and original commit:
    # https://github.com/open-contracting/lib-cove-ocds/blob/19ed9b3f0e392e9341206c5296d79c4bcc6f1206/libcoveocds/schema.py#L196-L235  # noqa: E501
    # https://github.com/OpenDataServices/cove/commit/1b03d44e3dcfa03313fb6b101075b9732c277ce2
    @functools.lru_cache()
    def get_pkg_schema_obj(self, deref=False, use_extensions=True):
        if hasattr(self, "_test_override_package_schema"):
            with open(self._test_override_package_schema) as f:
                if deref:
                    return jsonref.load(f, lazy_load=False)
                return json.load(f)
        else:
            schema = self._get_schema(self._package_schema_name)

        if deref:
            patched = self.get_schema_obj(deref=True)
            if self._package_schema_name == "record-package-schema.json":
                schema["definitions"]["record"]["properties"]["compiledRelease"] = patched
                schema["definitions"]["record"]["properties"]["releases"]["oneOf"][1]["items"] = patched
                schema = jsonref.replace_refs(schema, lazy_load=False)
            else:
                schema["properties"]["releases"]["items"] = patched

        return schema

    @functools.lru_cache()
    def get_schema_obj(self, deref=False):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", category=ExtensionWarning)

            schema = self.builder.patched_release_schema(schema=self._get_schema("release-schema.json"))
            if deref:
                try:
                    # Don't use proxies=False. If "$ref" has "deprecated" as a sibling, it is lost.
                    schema = jsonref.replace_refs(schema, lazy_load=False)
                except jsonref.JsonRefError as e:
                    # Callers must check json_deref_error.
                    self.json_deref_error = e.message
                    # This is the prior behavior, however surprising. https://github.com/OpenDataServices/lib-cove/blob/a97f7698d5e5ed519165e930a53fc97cc387362c/libcove/lib/common.py#L393-L405  # noqa: E501
                    schema = {}

            for warning in w:
                if issubclass(warning.category, ExtensionWarning):
                    exception = warning.message.exc
                    if isinstance(exception, requests.HTTPError):
                        message = f"{exception.response.status_code}: {exception.response.reason.lower()}"
                    elif isinstance(exception, (requests.RequestException, zipfile.BadZipFile)):
                        message = "fetching failed"
                    elif isinstance(exception, json.JSONDecodeError):
                        message = "release schema patch is not valid JSON"
                    else:
                        message = str(exception)
                    self.invalid_extension[warning.message.extension.input_url] = message

        language = self.config.config["current_language"]

        for extension in self.builder_extensions:
            input_url = extension.input_url

            # process_codelists() needs this dict.
            details = {"failed_codelists": {}}

            # Skip metadata fields in API context.
            if self.api:
                self.extensions[input_url] = details
                continue

            details["url"] = input_url
            details["schema_url"] = None

            # ocdsextensionregistry requires the input URL to contain "extension.json" (like the registry). If it
            # doesn't, ocdsextensionregistry can't determine how to retrieve it.
            if not extension.base_url and not extension._url_pattern:
                self.invalid_extension[input_url] = "missing extension.json"
                continue

            try:
                metadata = extension.metadata

                # We *could* check its existence via HTTP, but this is for display only anyway.
                schemas = metadata.get("schemas")
                if isinstance(schemas, list) and "release_schema.json" in schemas:
                    details["schema_url"] = extension.get_url("release-schema.json")

                for field in ("name", "description", "documentationUrl"):
                    language_map = metadata[field]
                    details[field] = language_map.get(language, language_map.get("en", ""))

                # In a web context, set the details only if metadata is available, otherwise cove-ocds' present
                # behavior to create an empty list item for the extension.
                self.extensions[input_url] = details
            except DoesNotExist:
                self.invalid_extension[input_url] = "404: not found"
            except requests.HTTPError as e:
                self.invalid_extension[input_url] = f"{e.response.status_code}: {e.response.reason.lower()}"
            except requests.RequestException:
                self.invalid_extension[input_url] = "fetching failed"
            except json.JSONDecodeError:
                self.invalid_extension[input_url] = "extension metadata is not valid JSON"

        # It's possible for the release schema to be applied, but the metadata file to be unavailable.
        # Nonetheless, for now, preserve prior behavior by reporting as if it were not applied.
        self.extended = len(self.invalid_extension) < len(self.extensions)

        return schema

    # This is always called with an `if schema_ocds.extensions` guard.
    def create_extended_schema_file(self, upload_dir, upload_url):
        basename = "extended_schema.json"
        path = os.path.join(upload_dir, basename)

        # Always replace any existing extended schema file
        if os.path.exists(path):
            os.remove(path)
            self.extended_schema_file = None
            self.extended_schema_url = None

        schema = self.get_schema_obj()
        if not self.extended:
            return

        with open(path, "w") as f:
            json.dump(schema, f, ensure_ascii=False, indent=2)
            f.write("\n")
        self.extended_schema_file = path
        self.extended_schema_url = urljoin(upload_url, basename)



import json

from libcoveocds.lib.additional_checks import CHECKS, run_additional_checks

checks = CHECKS["all"]


def test_empty_fields_releases_basic():
    with open("./tests/fixtures/additional_checks/empty_fields_releases.json", encoding="utf-8") as json_file:
        data = json.load(json_file)

    additional_checks = run_additional_checks(data, checks)
    result = {
        "empty_field": [
            {"json_location": "releases/0/parties/0/address"},
            {"json_location": "releases/0/planning/budget/id"},
            {"json_location": "releases/0/tender/items/0/additionalClassifications"},
        ]
    }

    assert additional_checks == result


def test_empty_fields_records_basic():
    with open("./tests/fixtures/additional_checks/empty_fields_records.json", encoding="utf-8") as json_file:
        data = json.load(json_file)

    additional_checks = run_additional_checks(data, checks)
    result = {
        "empty_field": [
            {"json_location": "records/0/compiledRelease/awards/0/documents/0/id"},
            {"json_location": "records/0/compiledRelease/awards/0/items/0/additionalClassifications"},
        ]
    }

    assert additional_checks == result


def test_empty_fields_empty_string():
    data = {
        "releases": [
            {
                "ocid": "ocds-213czf-000-00001",
                "date": "",
                "initiationType": "tender",
            }
        ]
    }

    assert run_additional_checks(data, checks) == {"empty_field": [{"json_location": "releases/0/date"}]}


def test_empty_fields_empty_dict():
    data = {"releases": [{"buyer": {}}]}

    assert run_additional_checks(data, checks) == {"empty_field": [{"json_location": "releases/0/buyer"}]}


def test_empty_fields_empty_list():
    data = {"releases": [{"parties": []}]}

    assert run_additional_checks(data, checks) == {"empty_field": [{"json_location": "releases/0/parties"}]}


def test_empty_fields_all_fine():
    with open("./tests/fixtures/additional_checks/basic_releases.json", encoding="utf-8") as json_file:
        data = json.load(json_file)

    assert run_additional_checks(data, checks) == {}

    with open("./tests/fixtures/additional_checks/full_record.json", encoding="utf-8") as json_file:
        data = json.load(json_file)

    assert run_additional_checks(data, checks) == {}



import copy
import json

import pytest
from libcove.lib.common import get_additional_codelist_values

import libcoveocds.config
import libcoveocds.schema
from tests import fixture_path

# Cache for faster tests.
config = libcoveocds.config.LibCoveOCDSConfig()
config.config["cache_all_requests"] = True

DEFAULT_OCDS_VERSION = libcoveocds.config.LIB_COVE_OCDS_CONFIG_DEFAULT["schema_version"]
METRICS_EXT = "https://raw.githubusercontent.com/open-contracting-extensions/ocds_metrics_extension/master/extension.json"  # noqa: E501
API_EXT = "https://chilecompracl.visualstudio.com/a6a3f587-5f23-42f6-9255-ac5852fae1e7/_apis/git/repositories/fb91c43b-011b-434b-901d-9d36ec50c586/items?path=%2Fextension.json&versionDescriptor%5BversionOptions%5D=0&versionDescriptor%5BversionType%5D=0&versionDescriptor%5Bversion%5D=master&resolveLfs=true&%24format=octetStream&api-version=5.0"  # noqa: E501
CODELIST_EXT = "https://raw.githubusercontent.com/INAImexico/ocds_extendedProcurementCategory_extension/0ed54770c85500cf21f46e88fb06a30a5a2132b1/extension.json"  # noqa: E501
UNSUPPORTED_PROTOCOL_EXT = "protocol://example.com/extension.json"
UNRESOLVABLE_HOST_EXT = "http://bad-url-for-extensions.com/extension.json"
NO_EXTENSION_EXT = "https://example.com/not-found"
OTHER_BASENAME_EXT = "https://example.com/not-found/other.json"
NO_FILES_EXT = "https://example.com/not-found/extension.json"
NO_METADATA_EXT = "https://raw.githubusercontent.com/open-contracting/lib-cove-ocds/main/tests/fixtures/extensions/no-metadata/extension.json"  # noqa: E501
INVALID_METADATA_EXT = "https://raw.githubusercontent.com/open-contracting/lib-cove-ocds/main/tests/fixtures/extensions/invalid-metadata/extension.json"  # noqa: E501


@pytest.mark.parametrize("record_pkg", [False, True])
def test_basic_1(record_pkg):
    schema = libcoveocds.schema.SchemaOCDS(record_pkg=record_pkg)

    assert schema.version == "1.1"
    assert schema.schema_url == "https://standard.open-contracting.org/1.1/en/release-schema.json"
    if record_pkg:
        assert schema.pkg_schema_url == "https://standard.open-contracting.org/1.1/en/record-package-schema.json"
    else:
        assert schema.pkg_schema_url == "https://standard.open-contracting.org/1.1/en/release-package-schema.json"
    assert schema.schema_host == "https://standard.open-contracting.org/1.1/en/"
    assert not schema.config.config["cache_all_requests"]


@pytest.mark.parametrize("record_pkg", [False, True])
def test_pass_config_1(record_pkg):

    config = copy.deepcopy(libcoveocds.config.LIB_COVE_OCDS_CONFIG_DEFAULT)
    config["schema_version"] = "1.0"

    lib_cove_ocds_config = libcoveocds.config.LibCoveOCDSConfig(config=config)

    schema = libcoveocds.schema.SchemaOCDS(lib_cove_ocds_config=lib_cove_ocds_config, record_pkg=record_pkg)

    assert schema.version == "1.0"
    assert schema.schema_url == "https://standard.open-contracting.org/1.0/en/release-schema.json"
    if record_pkg:
        assert schema.pkg_schema_url == "https://standard.open-contracting.org/1.0/en/record-package-schema.json"
    else:
        assert schema.pkg_schema_url == "https://standard.open-contracting.org/1.0/en/release-package-schema.json"
    assert schema.schema_host == "https://standard.open-contracting.org/1.0/en/"
    assert not schema.config.config["cache_all_requests"]


@pytest.mark.parametrize(
    ("select_version", "package_data", "version", "invalid_version_argument", "invalid_version_data", "extensions"),
    [
        (None, None, DEFAULT_OCDS_VERSION, False, False, {}),
        ("1.0", None, "1.0", False, False, {}),
        (None, {"version": "1.1"}, "1.1", False, False, {}),
        (None, {"version": "1.1", "extensions": ["c", "d"]}, "1.1", False, False, {"c": {}, "d": {}}),
        ("1.1", {"version": "1.0"}, "1.1", False, False, {}),
        ("1.bad", {"version": "1.1"}, "1.1", True, False, {}),
        ("1.wrong", {"version": "1.bad"}, DEFAULT_OCDS_VERSION, True, True, {}),
        (None, {"version": "1.bad"}, DEFAULT_OCDS_VERSION, False, True, {}),
        (None, {"extensions": ["a", "b"]}, "1.1", False, False, {"a": {}, "b": {}}),
        (None, {"version": "1.1", "extensions": ["a", "b"]}, "1.1", False, False, {"a": {}, "b": {}}),
    ],
)
def test_schema_ocds_constructor(
    select_version, package_data, version, invalid_version_argument, invalid_version_data, extensions
):
    schema = libcoveocds.schema.SchemaOCDS(select_version=select_version, package_data=package_data)
    host = libcoveocds.config.LIB_COVE_OCDS_CONFIG_DEFAULT["schema_version_choices"][version][1]
    url = f"{host}release-package-schema.json"

    assert schema.version == version
    assert schema.schema_host == host
    assert schema.pkg_schema_url == url
    assert schema.invalid_version_argument == invalid_version_argument
    assert schema.invalid_version_data == invalid_version_data
    assert schema.extensions == extensions


@pytest.mark.parametrize(
    ("package_data", "extensions", "invalid_extension", "extended", "extends_schema"),
    [
        (None, {}, {}, False, False),
        ({"version": "1.1", "extensions": [METRICS_EXT]}, {METRICS_EXT: {}}, {}, True, True),
        ({"version": "1.1", "extensions": [API_EXT]}, {API_EXT: {}}, {}, True, True),
        ({"version": "1.1", "extensions": [CODELIST_EXT]}, {CODELIST_EXT: {}}, {}, True, False),
        (
            {"version": "1.1", "extensions": [UNSUPPORTED_PROTOCOL_EXT]},
            {UNSUPPORTED_PROTOCOL_EXT: {}},
            {UNSUPPORTED_PROTOCOL_EXT: "fetching failed"},
            False,
            False,
        ),
        (
            {"version": "1.1", "extensions": [UNRESOLVABLE_HOST_EXT]},
            {UNRESOLVABLE_HOST_EXT: {}},
            {UNRESOLVABLE_HOST_EXT: "fetching failed"},
            False,
            False,
        ),
        (
            {"version": "1.1", "extensions": [NO_EXTENSION_EXT]},
            {NO_EXTENSION_EXT: {}},
            {NO_EXTENSION_EXT: "missing extension.json"},
            False,
            False,
        ),
        (
            {"version": "1.1", "extensions": [OTHER_BASENAME_EXT]},
            {OTHER_BASENAME_EXT: {}},
            {OTHER_BASENAME_EXT: "missing extension.json"},
            False,
            False,
        ),
        (
            {"version": "1.1", "extensions": [NO_FILES_EXT]},
            {NO_FILES_EXT: {}},
            {NO_FILES_EXT: "404: not found"},
            False,
            False,
        ),
        (
            {"version": "1.1", "extensions": [NO_METADATA_EXT]},
            {NO_METADATA_EXT: {}},
            {NO_METADATA_EXT: "404: not found"},
            False,
            False,
        ),
        (
            {"version": "1.1", "extensions": [INVALID_METADATA_EXT]},
            {INVALID_METADATA_EXT: {}},
            {INVALID_METADATA_EXT: "extension metadata is not valid JSON"},
            False,
            False,
        ),
        (
            {"version": "1.1", "extensions": [UNRESOLVABLE_HOST_EXT, METRICS_EXT]},
            {UNRESOLVABLE_HOST_EXT: {}, METRICS_EXT: {}},
            {UNRESOLVABLE_HOST_EXT: "fetching failed"},
            True,
            True,
        ),
    ],
)
def test_schema_ocds_extensions(package_data, extensions, invalid_extension, extended, extends_schema):
    schema = libcoveocds.schema.SchemaOCDS(package_data=package_data, lib_cove_ocds_config=config)
    assert schema.extensions == extensions
    assert not schema.extended

    schema_obj = schema.get_schema_obj()
    assert schema.invalid_extension == invalid_extension
    assert schema.extended is extended

    if extends_schema and METRICS_EXT in extensions:
        assert "Metric" in schema_obj["definitions"]
        assert "agreedMetrics" in schema_obj["definitions"]["Award"]["properties"]
    elif extends_schema and API_EXT in extensions:
        assert "dateCreated" in schema_obj["definitions"]["Contract"]["properties"]
    else:
        assert "Metric" not in schema_obj["definitions"]
        assert not schema_obj["definitions"]["Award"]["properties"].get("agreedMetrics")


# https://github.com/OpenDataServices/cove/issues/1054
@pytest.mark.xfail(reason="lib-cove has a bug")
def test_get_additional_codelist_values_replaced():
    with open(fixture_path("fixtures", "common_checks", "get_additional_codelist_values_replaced.json")) as f:
        package_data = json.load(f)

    # tariffs adds to documentType.csv, then ppp replaces documentType.csv, such that the addition has no consequence.
    # lib-cove's get_additional_codelist_values() determines the correct extension, but the incorrect filename.

    schema_obj = libcoveocds.schema.SchemaOCDS(select_version="1.1", package_data=package_data)
    schema_obj.get_schema_obj()

    additional_codelist_values = get_additional_codelist_values(schema_obj, package_data)

    assert additional_codelist_values["releases/tender/documents/documentType"] == {
        "field": "documentType",
        "codelist": "documentType.csv",
        "path": "releases/tender/documents",
        "codelist_url": "https://raw.githubusercontent.com/open-contracting-extensions/ocds_ppp_extension/master/codelists/documentType.csv",  # noqa
        "extension_codelist": False,
        "isopen": True,
        "codelist_amend_urls": [],
        "values": ["foo"],
    }



from collections import defaultdict

import libcove.lib.tools as tools


def flatten_dict(data, path=""):
    for key, value in data.items():
        if not value:
            yield (f"{path}/{key}", value)
        elif isinstance(value, dict):
            yield from flatten_dict(value, f"{path}/{key}")
        elif isinstance(value, list):
            for i, item in enumerate(value):
                if isinstance(item, dict):
                    yield from flatten_dict(item, f"{path}/{key}/{i}")
                else:
                    yield (f"{path}/{key}/{i}", item)
        else:
            yield (f"{path}/{key}", value)


def empty_field(data, path_prefix):
    """Identifying when fields, objects and arrays exist but are empty or contain only whitespace"""

    for key, value in flatten_dict(data):
        if isinstance(value, str) and not value.strip() or isinstance(value, (dict, list)) and not value:
            yield {"json_location": f"{path_prefix}{key}"}


CHECKS = {"all": [empty_field], "none": []}


@tools.ignore_errors
def run_additional_checks(data, functions):
    if "records" in data:
        key = "records"
    elif "releases" in data:
        key = "releases"
    else:
        return []

    results = defaultdict(list)

    for i, data in enumerate(data[key]):
        for function in functions:
            for output in function(data, f"{key}/{i}"):
                results[function.__name__].append(output)

    # https://stackoverflow.com/a/12842716/244258
    results.default_factory = None
    return results



import json
import os
import shutil
import tempfile

import pytest

import libcoveocds.common_checks
import libcoveocds.config
import libcoveocds.schema
from tests import fixture_path

# Cache for faster tests.
config = libcoveocds.config.LibCoveOCDSConfig()
config.config["cache_all_requests"] = True


def test_basic_1():

    cove_temp_folder = tempfile.mkdtemp(prefix="libcoveocds-tests-", dir=tempfile.gettempdir())
    schema = libcoveocds.schema.SchemaOCDS(lib_cove_ocds_config=config)
    json_filename = fixture_path("fixtures", "common_checks", "basic_1.json")
    with open(json_filename) as fp:
        json_data = json.load(fp)

    context = {
        "file_type": "json",
    }

    try:

        results = libcoveocds.common_checks.common_checks_ocds(
            context,
            cove_temp_folder,
            json_data,
            schema,
        )

    finally:
        shutil.rmtree(cove_temp_folder)

    assert results["version_used"] == "1.1"


def test_dupe_ids_1():

    cove_temp_folder = tempfile.mkdtemp(prefix="libcoveocds-tests-", dir=tempfile.gettempdir())
    schema = libcoveocds.schema.SchemaOCDS(lib_cove_ocds_config=config)
    json_filename = fixture_path("fixtures", "common_checks", "dupe_ids_1.json")
    with open(json_filename) as fp:
        json_data = json.load(fp)

    context = {
        "file_type": "json",
    }

    try:

        results = libcoveocds.common_checks.common_checks_ocds(
            context,
            cove_temp_folder,
            json_data,
            schema,
        )

    finally:
        shutil.rmtree(cove_temp_folder)

    # https://github.com/OpenDataServices/cove/issues/782 Defines how we want this error shown
    assert len(results["validation_errors"][0][1]) == 2
    # test paths
    assert results["validation_errors"][0][1][0]["path"] == "releases"
    assert results["validation_errors"][0][1][1]["path"] == "releases"
    # test values
    # we don't know what order they will come out in, so fix the order ourselves
    values = [
        results["validation_errors"][0][1][0]["value"],
        results["validation_errors"][0][1][1]["value"],
    ]
    values.sort()
    assert values[0] == "ocds-213czf-000-00001, ocds-213czf-000-00001-01-planning"
    assert values[1] == "ocds-213czf-000-00001, ocds-213czf-000-00001-02-planning"


@pytest.mark.parametrize(
    "record_pkg,filename,schema_subdir,validation_error_jsons_expected",
    [
        (False, "releases_no_validation_errors.json", "", []),
        (True, "records_no_validation_errors.json", "", []),
        (
            True,
            "records_invalid_releases.json",
            "",
            [
                {
                    "message": "'date' is missing but required within 'releases'",
                    "message_safe": "&#x27;date&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "date",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'date' is missing but required within 'releases'",
                    "message_safe": "&#x27;date&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "linked_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "date",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [
                        {"path": "records/1/releases/0"},
                        {"path": "records/3/releases/0"},
                    ],
                },
                {
                    "message": "'initiationType' is missing but required within 'releases'",
                    "message_safe": "&#x27;initiationType&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "initiationType",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'ocid' is missing but required within 'releases'",
                    "message_safe": "&#x27;ocid&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "ocid",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'releases' is not a JSON array",
                    "message_safe": "&#x27;releases&#x27; is not a JSON array",
                    "validator": "type",
                    "assumption": "linked_releases",
                    "message_type": "array",
                    "path_no_number": "records/releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "is not null, and",
                    "error_id": None,
                    "values": [
                        {"path": "records/6/releases", "value": "a string"},
                        {"path": "records/7/releases", "value": None},
                        {"path": "records/8/releases"},
                    ],
                },
                {
                    "message": "'tag' is missing but required within 'releases'",
                    "message_safe": "&#x27;tag&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "tag",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'url' is missing but required within 'releases'",
                    "message_safe": "&#x27;url&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "linked_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "url",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/1/releases/0"}],
                },
                {
                    "message": "This array should contain either entirely embedded releases or linked releases. Embedded releases contain an 'id' whereas linked releases do not. Your releases contain a mixture.",  # noqa: E501
                    "message_safe": "This array should contain either entirely embedded releases or linked releases. Embedded releases contain an &#x27;id&#x27; whereas linked releases do not. Your releases contain a mixture.",  # noqa: E501
                    "validator": "oneOf",
                    "assumption": None,
                    "message_type": "oneOf",
                    "path_no_number": "records/releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "",
                    "error_id": "releases_both_embedded_and_linked",
                    "values": [
                        {"path": "records/4/releases"},
                        {"path": "records/5/releases"},
                    ],
                },
                {
                    "message": "[] is too short",
                    "message_safe": "[] is too short",
                    "validator": "minItems",
                    "assumption": "linked_releases",
                    "message_type": "minItems",
                    "path_no_number": "records/releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/0/releases"}],
                    "instance": [],
                },
            ],
        ),
        (
            True,
            "records_invalid_releases.json",
            "1-0",
            [
                {
                    "message": "'date' is missing but required within 'releases'",
                    "message_safe": "&#x27;date&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "date",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'date' is missing but required within 'releases'",
                    "message_safe": "&#x27;date&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "linked_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "date",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [
                        {"path": "records/1/releases/0"},
                        {"path": "records/3/releases/0"},
                    ],
                },
                {
                    "message": "'initiationType' is missing but required within 'releases'",
                    "message_safe": "&#x27;initiationType&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "initiationType",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'ocid' is missing but required within 'releases'",
                    "message_safe": "&#x27;ocid&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "ocid",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'releases' is not a JSON array",
                    "message_safe": "&#x27;releases&#x27; is not a JSON array",
                    "validator": "type",
                    "assumption": "linked_releases",
                    "message_type": "array",
                    "path_no_number": "records/releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "is not null, and",
                    "error_id": None,
                    "values": [
                        {"path": "records/6/releases", "value": "a string"},
                        {"path": "records/7/releases", "value": None},
                        {"path": "records/8/releases"},
                    ],
                },
                {
                    "message": "'tag' is missing but required within 'releases'",
                    "message_safe": "&#x27;tag&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "embedded_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "tag",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/2/releases/0"}],
                },
                {
                    "message": "'url' is missing but required within 'releases'",
                    "message_safe": "&#x27;url&#x27; is missing but required within &#x27;releases&#x27;",
                    "validator": "required",
                    "assumption": "linked_releases",
                    "message_type": "required",
                    "path_no_number": "records/releases",
                    "header": "url",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/1/releases/0"}],
                },
                {
                    "message": "This array should contain either entirely embedded releases or linked releases. Embedded releases contain an 'id' whereas linked releases do not. Your releases contain a mixture.",  # noqa: E501
                    "message_safe": "This array should contain either entirely embedded releases or linked releases. Embedded releases contain an &#x27;id&#x27; whereas linked releases do not. Your releases contain a mixture.",  # noqa: E501
                    "validator": "oneOf",
                    "assumption": None,
                    "message_type": "oneOf",
                    "path_no_number": "records/releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "",
                    "error_id": "releases_both_embedded_and_linked",
                    "values": [
                        {"path": "records/4/releases"},
                        {"path": "records/5/releases"},
                    ],
                },
                {
                    "message": "[] is too short",
                    "message_safe": "[] is too short",
                    "validator": "minItems",
                    "assumption": "linked_releases",
                    "message_type": "minItems",
                    "path_no_number": "records/releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/0/releases"}],
                    "instance": [],
                },
            ],
        ),
        (
            False,
            "releases_non_unique.json",
            "",
            [
                {
                    "message": "Non-unique combination of ocid, id values",
                    "message_safe": "Non-unique combination of ocid, id values",
                    "validator": "uniqueItems",
                    "assumption": None,
                    "message_type": "uniqueItems",
                    "path_no_number": "releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "",
                    "error_id": "uniqueItems_with_ocid__id",
                    "values": [
                        {"path": "releases", "value": "EXAMPLE-1, EXAMPLE-1-1"},
                        {"path": "releases", "value": "EXAMPLE-1, EXAMPLE-1-2"},
                    ],
                }
            ],
        ),
        (False, "releases_unique_ocids_but_not_ids.json", "", []),
        (
            True,
            "records_non_unique.json",
            "",
            [
                {
                    "message": "Non-unique ocid values",
                    "message_safe": "Non-unique ocid values",
                    "validator": "uniqueItems",
                    "assumption": None,
                    "message_type": "uniqueItems",
                    "path_no_number": "records",
                    "header": "records",
                    "header_extra": "records",
                    "null_clause": "",
                    "error_id": "uniqueItems_with_ocid",
                    "values": [
                        {"path": "records", "value": "EXAMPLE-1"},
                        {"path": "records", "value": "EXAMPLE-2"},
                    ],
                }
            ],
        ),
        (
            False,
            "releases_non_unique_no_id.json",
            "",
            [
                {
                    "message": "'id' is missing but required",
                    "message_safe": "&#x27;id&#x27; is missing but required",
                    "validator": "required",
                    "assumption": None,
                    "message_type": "required",
                    "path_no_number": "releases",
                    "header": "id",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "releases/0"}, {"path": "releases/1"}],
                },
                {
                    "message": "Array has non-unique elements",
                    "message_safe": "Array has non-unique elements",
                    "validator": "uniqueItems",
                    "assumption": None,
                    "message_type": "uniqueItems",
                    "path_no_number": "releases",
                    "header": "releases",
                    "header_extra": "releases",
                    "null_clause": "",
                    "error_id": "uniqueItems_no_ids",
                    "values": [{"path": "releases"}],
                },
            ],
        ),
        (
            True,
            "records_non_unique_no_ocid.json",
            "",
            [
                {
                    "message": "'ocid' is missing but required",
                    "message_safe": "&#x27;ocid&#x27; is missing but required",
                    "validator": "required",
                    "assumption": None,
                    "message_type": "required",
                    "path_no_number": "records",
                    "header": "ocid",
                    "header_extra": "records/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/0"}, {"path": "records/1"}],
                },
                {
                    "message": "Array has non-unique elements",
                    "message_safe": "Array has non-unique elements",
                    "validator": "uniqueItems",
                    "assumption": None,
                    "message_type": "uniqueItems",
                    "path_no_number": "records",
                    "header": "records",
                    "header_extra": "records",
                    "null_clause": "",
                    "error_id": "uniqueItems_no_ids",
                    "values": [{"path": "records"}],
                },
            ],
        ),
        # Check that we handle unique arrays correctly also
        # (e.g. that we don't incorrectly claim they are not unique)
        (
            False,
            "releases_unique.json",
            "",
            [
                {
                    "message": "'id' is missing but required",
                    "message_safe": "&#x27;id&#x27; is missing but required",
                    "validator": "required",
                    "assumption": None,
                    "message_type": "required",
                    "path_no_number": "releases",
                    "header": "id",
                    "header_extra": "releases/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "releases/0"}, {"path": "releases/1"}],
                }
            ],
        ),
        (
            True,
            "records_unique.json",
            "",
            [
                {
                    "message": "'ocid' is missing but required",
                    "message_safe": "&#x27;ocid&#x27; is missing but required",
                    "validator": "required",
                    "assumption": None,
                    "message_type": "required",
                    "path_no_number": "records",
                    "header": "ocid",
                    "header_extra": "records/[number]",
                    "null_clause": "",
                    "error_id": None,
                    "values": [{"path": "records/0"}, {"path": "records/1"}],
                }
            ],
        ),
    ],
)
def test_validation_release_or_record_package(record_pkg, filename, validation_error_jsons_expected, schema_subdir):
    schema_host = fixture_path("fixtures", "common_checks", schema_subdir, "")
    with open(os.path.join(schema_host, filename)) as fp:
        json_data = json.load(fp)

    cove_temp_folder = tempfile.mkdtemp(prefix="libcoveocds-tests-", dir=tempfile.gettempdir())
    schema = libcoveocds.schema.SchemaOCDS(lib_cove_ocds_config=config, record_pkg=record_pkg)
    context = {
        "file_type": "json",
    }

    results = libcoveocds.common_checks.common_checks_ocds(
        context,
        cove_temp_folder,
        json_data,
        schema,
    )

    validation_errors = results["validation_errors"]

    validation_error_jsons = []
    for validation_error_json, values in validation_errors:
        validation_error_json = json.loads(validation_error_json)
        validation_error_json["values"] = values
        # Remove this as it can be a rather large schema object
        del validation_error_json["validator_value"]
        # TODO: these are removed as they were't present in the lib-cove JSON
        # We should maybe add them to the validation error fixtures now
        if "docs_ref" in validation_error_json:
            del validation_error_json["docs_ref"]
        if "schema_description_safe" in validation_error_json:
            del validation_error_json["schema_description_safe"]
        if "schema_title" in validation_error_json:
            del validation_error_json["schema_title"]
        validation_error_jsons.append(validation_error_json)

    def strip_nones(list_of_dicts):
        out = []
        for a_dict in list_of_dicts:
            out.append({key: value for key, value in a_dict.items() if value is not None})
        return out

    assert strip_nones(validation_error_jsons) == strip_nones(validation_error_jsons_expected)


def test_ref_error(tmpdir):
    url = "https://raw.githubusercontent.com/open-contracting/lib-cove-ocds/main/tests/fixtures/extensions/unresolvable/extension.json"  # noqa: E501

    json_data = {"version": "1.1", "extensions": [url], "releases": [{"unresolvable": "1"}]}

    schema = libcoveocds.schema.SchemaOCDS("1.1", json_data, lib_cove_ocds_config=config)
    schema.create_extended_schema_file(tmpdir, "")

    libcoveocds.common_checks.common_checks_ocds(
        {"file_type": "json"},
        tmpdir,
        json_data,
        schema,
    )



import collections
import re

import libcove.lib.tools as tools


@tools.ignore_errors
def get_releases_aggregates(json_data):
    release_count = 0
    unique_ocids = set()
    tags = collections.Counter()
    unique_lang = set()
    unique_initation_type = set()
    unique_release_ids = set()
    duplicate_release_ids = set()

    #  for matching with contracts
    unique_award_id = set()

    planning_ocids = set()
    tender_ocids = set()
    awardid_ocids = set()
    award_ocids = set()
    contractid_ocids = set()
    contract_ocids = set()
    implementation_contractid_ocids = set()
    implementation_ocids = set()

    release_dates = []
    tender_dates = []
    award_dates = []
    contract_dates = []

    unique_buyers_identifier = {}
    unique_buyers_name_no_id = set()
    unique_suppliers_identifier = {}
    unique_suppliers_name_no_id = set()
    unique_procuring_identifier = {}
    unique_procuring_name_no_id = set()
    unique_tenderers_identifier = {}
    unique_tenderers_name_no_id = set()

    unique_organisation_schemes = set()
    organisation_identifier_address = set()
    organisation_name_no_id_address = set()
    organisation_identifier_contact_point = set()
    organisation_name_no_id_contact_point = set()

    release_tender_item_ids = set()
    release_award_item_ids = set()
    release_contract_item_ids = set()
    item_identifier_schemes = set()
    unique_item_ids = set()

    unique_currency = set()

    planning_doctype = collections.Counter()
    planning_doc_count = 0
    tender_doctype = collections.Counter()
    tender_doc_count = 0
    tender_milestones_doctype = collections.Counter()
    tender_milestones_doc_count = 0
    award_doctype = collections.Counter()
    award_doc_count = 0
    contract_doctype = collections.Counter()
    contract_doc_count = 0
    implementation_doctype = collections.Counter()
    implementation_doc_count = 0
    implementation_milestones_doctype = collections.Counter()
    implementation_milestones_doc_count = 0

    def process_org(org, unique_id, unique_name):
        if not isinstance(org, dict):
            return

        identifier = org.get("identifier")
        org_id = None
        if identifier:
            org_id = identifier.get("id")
            if org_id:
                unique_id[org_id] = org.get("name", "") or ""
                scheme = identifier.get("scheme")
                if scheme:
                    unique_organisation_schemes.add(scheme)
                if org.get("address"):
                    organisation_identifier_address.add(org_id)
                if org.get("contactPoint"):
                    organisation_identifier_contact_point.add(org_id)
        if not org_id:
            name = org.get("name")
            if name:
                unique_name.add(name)
            if org.get("address"):
                organisation_name_no_id_address.add(name)
            if org.get("contactPoint"):
                organisation_name_no_id_contact_point.add(name)

    def get_item_scheme(item):
        classification = item.get("classification")
        if classification:
            scheme = classification.get("scheme")
            if scheme:
                item_identifier_schemes.add(scheme)

    releases = tools.get_no_exception(json_data, "releases", [])
    for release in releases:
        # ### Release Section ###
        release_count = release_count + 1
        ocid = release.get("ocid")
        release_id = release.get("id")
        if not ocid:
            continue
        if release_id:
            if release_id in unique_release_ids:
                duplicate_release_ids.add(release_id)
            unique_release_ids.add(release_id)

        unique_ocids.add(release["ocid"])
        if "tag" in release:
            tags.update(tools.to_list(release["tag"]))
        initiation_type = release.get("initiationType")
        if initiation_type:
            unique_initation_type.add(initiation_type)

        release_date = release.get("date", "")
        if release_date:
            release_dates.append(str(release_date))

        if "language" in release:
            unique_lang.add(release["language"])
        buyer = release.get("buyer")
        if buyer:
            process_org(buyer, unique_buyers_identifier, unique_buyers_name_no_id)

        # ### Planning Section ###
        planning = tools.get_no_exception(release, "planning", {})
        if planning and isinstance(planning, dict):
            planning_ocids.add(ocid)
            planning_doc_count += tools.update_docs(planning, planning_doctype)

        # ### Tender Section ###
        tender = tools.get_no_exception(release, "tender", {})
        if tender and isinstance(tender, dict):
            tender_ocids.add(ocid)
            tender_doc_count += tools.update_docs(tender, tender_doctype)
            tender_period = tender.get("tenderPeriod")
            if tender_period:
                start_date = tender_period.get("startDate", "")
                if start_date:
                    tender_dates.append(str(start_date))
            procuring_entity = tender.get("procuringEntity")
            if procuring_entity:
                process_org(procuring_entity, unique_procuring_identifier, unique_procuring_name_no_id)
            tenderers = tender.get("tenderers", [])
            for tenderer in tenderers:
                process_org(tenderer, unique_tenderers_identifier, unique_tenderers_name_no_id)
            tender_items = tender.get("items", [])
            for item in tender_items:
                item_id = item.get("id")
                if item_id:
                    unique_item_ids.add(item_id)
                if item_id and release_id:
                    release_tender_item_ids.add((ocid, release_id, item_id))
                get_item_scheme(item)
            milestones = tender.get("milestones")
            if milestones:
                for milestone in milestones:
                    tender_milestones_doc_count += tools.update_docs(milestone, tender_milestones_doctype)

        # ### Award Section ###
        awards = tools.get_no_exception(release, "awards", [])
        for award in awards:
            if not isinstance(award, dict):
                continue
            award_id = award.get("id")
            award_ocids.add(ocid)
            if award_id:
                unique_award_id.add(award_id)
                awardid_ocids.add((award_id, ocid))
            award_date = award.get("date", "")
            if award_date:
                award_dates.append(str(award_date))
            award_items = award.get("items", [])
            for item in award_items:
                item_id = item.get("id")
                if item_id:
                    unique_item_ids.add(item_id)
                if item_id and release_id and award_id:
                    release_award_item_ids.add((ocid, release_id, award_id, item_id))
                get_item_scheme(item)
            suppliers = award.get("suppliers", [])
            for supplier in suppliers:
                process_org(supplier, unique_suppliers_identifier, unique_suppliers_name_no_id)
            award_doc_count += tools.update_docs(award, award_doctype)

        # ### Contract section
        contracts = tools.get_no_exception(release, "contracts", [])
        for contract in contracts:
            contract_id = contract.get("id")
            contract_ocids.add(ocid)
            if contract_id:
                contractid_ocids.add((contract_id, ocid))
            period = contract.get("period")
            if period:
                start_date = period.get("startDate", "")
                if start_date:
                    contract_dates.append(start_date)
            contract_items = contract.get("items", [])
            for item in contract_items:
                item_id = item.get("id")
                if item_id:
                    unique_item_ids.add(item_id)
                if item_id and release_id and contract_id:
                    release_contract_item_ids.add((ocid, release_id, contract_id, item_id))
                get_item_scheme(item)
            contract_doc_count += tools.update_docs(contract, contract_doctype)
            implementation = contract.get("implementation")
            if implementation:
                implementation_ocids.add(ocid)
                if contract_id:
                    implementation_contractid_ocids.add((contract_id, ocid))
                implementation_doc_count += tools.update_docs(implementation, implementation_doctype)
                implementation_milestones = implementation.get("milestones", [])
                for milestone in implementation_milestones:
                    implementation_milestones_doc_count += tools.update_docs(
                        milestone, implementation_milestones_doctype
                    )

    contracts_without_awards = []
    for release in releases:
        contracts = release.get("contracts", [])
        for contract in contracts:
            award_id = contract.get("awardID")
            if award_id not in unique_award_id:
                contracts_without_awards.append(contract)

    unique_buyers_count = len(unique_buyers_identifier) + len(unique_buyers_name_no_id)
    unique_buyers = [name + " (" + str(id) + ")" for id, name in unique_buyers_identifier.items()] + list(
        unique_buyers_name_no_id
    )

    unique_suppliers_count = len(unique_suppliers_identifier) + len(unique_suppliers_name_no_id)
    unique_suppliers = [name + " (" + str(id) + ")" for id, name in unique_suppliers_identifier.items()] + list(
        unique_suppliers_name_no_id
    )

    unique_procuring_count = len(unique_procuring_identifier) + len(unique_procuring_name_no_id)
    unique_procuring = [name + " (" + str(id) + ")" for id, name in unique_procuring_identifier.items()] + list(
        unique_procuring_name_no_id
    )

    unique_tenderers_count = len(unique_tenderers_identifier) + len(unique_tenderers_name_no_id)
    unique_tenderers = [name + " (" + str(id) + ")" for id, name in unique_tenderers_identifier.items()] + list(
        unique_tenderers_name_no_id
    )

    unique_org_identifier_count = len(
        set(unique_buyers_identifier)
        | set(unique_suppliers_identifier)
        | set(unique_procuring_identifier)
        | set(unique_tenderers_identifier)
    )
    unique_org_name_count = len(
        unique_buyers_name_no_id
        | unique_suppliers_name_no_id
        | unique_procuring_name_no_id
        | unique_tenderers_name_no_id
    )
    unique_org_count = unique_org_identifier_count + unique_org_name_count

    def get_currencies(object):
        if isinstance(object, dict):
            for key, value in object.items():
                if key == "currency":
                    unique_currency.add(value)
                get_currencies(value)
        if isinstance(object, list):
            for item in object:
                get_currencies(item)

    get_currencies(json_data)

    return {
        "release_count": release_count,
        "unique_ocids": sorted(unique_ocids, key=lambda x: str(x)),
        "unique_initation_type": sorted(unique_initation_type, key=lambda x: str(x)),
        "duplicate_release_ids": sorted(duplicate_release_ids, key=lambda x: str(x)),
        "tags": dict(tags),
        "unique_lang": sorted(unique_lang, key=lambda x: str(x)),
        "unique_award_id": sorted(unique_award_id, key=lambda x: str(x)),
        "planning_count": len(planning_ocids),
        "tender_count": len(tender_ocids),
        "award_count": len(awardid_ocids),
        "processes_award_count": len(award_ocids),
        "contract_count": len(contractid_ocids),
        "processes_contract_count": len(contract_ocids),
        "implementation_count": len(implementation_contractid_ocids),
        "processes_implementation_count": len(implementation_ocids),
        "min_release_date": min(release_dates) if release_dates else "",
        "max_release_date": max(release_dates) if release_dates else "",
        "min_tender_date": min(tender_dates) if tender_dates else "",
        "max_tender_date": max(tender_dates) if tender_dates else "",
        "min_award_date": min(award_dates) if award_dates else "",
        "max_award_date": max(award_dates) if award_dates else "",
        "min_contract_date": min(contract_dates) if contract_dates else "",
        "max_contract_date": max(contract_dates) if contract_dates else "",
        "unique_buyers_identifier": unique_buyers_identifier,
        "unique_buyers_name_no_id": sorted(unique_buyers_name_no_id, key=lambda x: str(x)),
        "unique_suppliers_identifier": unique_suppliers_identifier,
        "unique_suppliers_name_no_id": sorted(unique_suppliers_name_no_id, key=lambda x: str(x)),
        "unique_procuring_identifier": unique_procuring_identifier,
        "unique_procuring_name_no_id": sorted(unique_procuring_name_no_id, key=lambda x: str(x)),
        "unique_tenderers_identifier": unique_tenderers_identifier,
        "unique_tenderers_name_no_id": sorted(unique_tenderers_name_no_id, key=lambda x: str(x)),
        "unique_buyers": sorted(set(unique_buyers)),
        "unique_suppliers": sorted(set(unique_suppliers)),
        "unique_procuring": sorted(set(unique_procuring)),
        "unique_tenderers": sorted(set(unique_tenderers)),
        "unique_buyers_count": unique_buyers_count,
        "unique_suppliers_count": unique_suppliers_count,
        "unique_procuring_count": unique_procuring_count,
        "unique_tenderers_count": unique_tenderers_count,
        "unique_org_identifier_count": unique_org_identifier_count,
        "unique_org_name_count": unique_org_name_count,
        "unique_org_count": unique_org_count,
        "unique_organisation_schemes": sorted(unique_organisation_schemes, key=lambda x: str(x)),
        "organisations_with_address": len(organisation_identifier_address) + len(organisation_name_no_id_address),
        "organisations_with_contact_point": len(organisation_identifier_contact_point)
        + len(organisation_name_no_id_contact_point),
        "total_item_count": len(release_tender_item_ids)
        + len(release_award_item_ids)
        + len(release_contract_item_ids),
        "tender_item_count": len(release_tender_item_ids),
        "award_item_count": len(release_award_item_ids),
        "contract_item_count": len(release_contract_item_ids),
        "unique_item_ids_count": len(unique_item_ids),
        "item_identifier_schemes": sorted(item_identifier_schemes, key=lambda x: str(x)),
        "unique_currency": sorted(unique_currency, key=lambda x: str(x)),
        "planning_doc_count": planning_doc_count,
        "tender_doc_count": tender_doc_count,
        "tender_milestones_doc_count": tender_milestones_doc_count,
        "award_doc_count": award_doc_count,
        "contract_doc_count": contract_doc_count,
        "implementation_doc_count": implementation_doc_count,
        "implementation_milestones_doc_count": implementation_milestones_doc_count,
        "planning_doctype": dict(planning_doctype),
        "tender_doctype": dict(tender_doctype),
        "tender_milestones_doctype": dict(tender_milestones_doctype),
        "award_doctype": dict(award_doctype),
        "contract_doctype": dict(contract_doctype),
        "implementation_doctype": dict(implementation_doctype),
        "implementation_milestones_doctype": dict(implementation_milestones_doctype),
        "contracts_without_awards": contracts_without_awards,
    }


@tools.ignore_errors
def get_records_aggregates(json_data):
    # Unique ocids
    unique_ocids = set()

    if "records" in json_data:
        for record in json_data["records"]:
            # Gather all the ocids
            if "ocid" in record:
                unique_ocids.add(record["ocid"])

    # Number of records
    count = len(json_data["records"]) if "records" in json_data else 0

    return {
        "count": count,
        "unique_ocids": unique_ocids,
    }


def get_bad_ocid_prefixes(json_data):
    """Yield tuples with ('ocid', 'path/to/ocid') for ocids with malformed prefixes"""
    if not isinstance(json_data, dict):
        return []

    prefix_regex = re.compile(r"^ocds-[a-z0-9]{6}")

    def _is_bad_prefix(item):
        if (
            isinstance(item, dict)
            and (ocid := item.get("ocid"))
            and isinstance(ocid, str)
            and not prefix_regex.match(ocid)
        ):
            return ocid

    records = json_data.get("records")
    if records:
        bad_prefixes = []
        if isinstance(records, list):
            for i, record in enumerate(records):
                if not isinstance(record, dict):
                    continue

                if ocid := _is_bad_prefix(record):
                    bad_prefixes.append((ocid, f"records/{i}/ocid"))

                releases = record.get("releases")
                if isinstance(releases, list):
                    for j, release in enumerate(releases):
                        if ocid := _is_bad_prefix(release):
                            bad_prefixes.append((ocid, f"records/{i}/releases/{j}/ocid"))

                compiled_release = record.get("compiledRelease")
                if ocid := _is_bad_prefix(compiled_release):
                    bad_prefixes.append((ocid, f"records/{i}/compiledRelease/ocid"))
        return bad_prefixes

    releases = json_data.get("releases")
    if releases:
        bad_prefixes = []
        if isinstance(releases, list):
            for j, release in enumerate(releases):
                if ocid := _is_bad_prefix(release):
                    bad_prefixes.append((ocid, f"releases/{j}/ocid"))
        return bad_prefixes

    return []
